@@ -89,7 +89,7 @@ class RestapiEncoder(json.JSONEncoder):
     def default(self, o):
         """Encodes object for JSON.
 
-        Arg:
+        Args:
             o: Object.
         """
         if o == True:
@@ -136,8 +136,8 @@ class IdentityManager(object):
         """Returns a token for a specific domain from token store if one has been
                 generated for the ArcGIS Server resource.
 
-        Arg:
-            url: URL for secured resource.
+        Args:
+            url: URL for secured resource, or token as a string.
 
         Raises:
             TokenExpired: 'Token expired at {}! Please sign in again.'
@@ -152,7 +152,7 @@ class IdentityManager(object):
             #     url = url.split('/rest/services')[0] + '/rest/services'
             to_remove = []
             for registered_url, token in list(self.tokens.items()) + list(self._portal_tokens.items()):
-                if fnmatch.fnmatch(url, registered_url + '*'):
+                if fnmatch.fnmatch(url, registered_url + '*') or token['token'] == url:
                     if not token.isExpired:
                         return token
                     else:
@@ -174,7 +174,7 @@ class IdentityManager(object):
         """Returns a proxy url for a specific domain from token store if one has been
                 used to access the ArcGIS Server resource
 
-        Arg:
+        Args:
             url: URL for secured resource.
         """
 
@@ -185,6 +185,25 @@ class IdentityManager(object):
 
         return None
 
+
+    def flush(self, expired_only=True):
+        """Flush expired or all tokens/proxies from the Identity Manager.
+
+        Args:
+            expired_only (bool): If True, will only remove expired tokens. If
+                False, will remove all tokens and proxies. Default is True.
+
+        """
+        if expired_only:
+            for tokens in (self.tokens, self._portal_tokens):
+                for url in list(tokens.keys()):
+                    token = tokens[url]
+                    if token.isExpired:
+                        del tokens[url]
+        else:
+            self.__init__()
+
+
 # initialize Identity Manager
 ID_MANAGER = IdentityManager()
 
@@ -250,7 +269,7 @@ def tmp_json_file():
         TEMP_DIR = tempfile.mkdtemp()
     return os.path.join(TEMP_DIR, 'restapi_{}.json'.format(time.strftime('%Y%m%d%H%M%S')))
 
-def do_request(service, params={F: JSON}, ret_json=True, token='', cookies=None, proxy=None, referer=None, client=None, **kwargs):
+def do_request(service, params={F: JSON}, ret_json=True, token='', cookies=None, proxy=None, referer=None, client=None, method='get', **kwargs):
     """Post Request to REST Endpoint through query string, to post
             request with data in body, use requests.post(url, data:{k : v}).
 
@@ -275,6 +294,7 @@ def do_request(service, params={F: JSON}, ret_json=True, token='', cookies=None,
     Returns:
         The post request.
     """
+    ID_MANAGER.flush()
     global PROTOCOL
     if PROTOCOL != '':
         service = '{}://{}'.format(PROTOCOL, service.split('://')[-1])
@@ -328,6 +348,20 @@ def do_request(service, params={F: JSON}, ret_json=True, token='', cookies=None,
 
     stream = params.get(F) == 'image'
 
+    # handle cookies specially to merge
+    kwarg_cookies = kwargs.get('cookies')
+    if isinstance(kwarg_cookies, dict):
+        kwarg_cookies.update(cookies)
+    else:
+        kwargs['cookies'] = cookies
+
+    # if using stream, we probably want attachment, need to disable format
+    if kwargs.get('stream'):
+        if params[F]:
+            del params[F]
+        stream = True
+        ret_json = False
+
     # mixin default kwargs for requests
     defaults = {
         "stream": stream,
@@ -336,21 +370,13 @@ def do_request(service, params={F: JSON}, ret_json=True, token='', cookies=None,
         if k not in kwargs:
             kwargs[k] = v
 
-    # handle cookies specially to merge
-    kwarg_cookies = kwargs.get('cookies')
-    if isinstance(kwarg_cookies, dict):
-        kwarg_cookies.update(cookies)
-    else:
-        kwargs['cookies'] = cookies
-
     if proxy:
         # IMPORTANT: this is not a regular proxy, this is the Esri Proxy
         # see: https://github.com/Esri/resource-proxy
         r = do_proxy_request(proxy, service, params, referer, client=client)
         ID_MANAGER.proxies[service.split('/rest')[0].lower() + '/rest/services'] = proxy
     else:
-        request_method = get_request_method(service, params, client=client, method=kwargs.get('method', 'get'))
-        kwargs.pop('method', None)
+        request_method = get_request_method(service, params, client=client, method=method)
         if request_method.__name__ == 'get':
             # must use kwargs after url in GET
             r = request_method(service, params=params, **kwargs)
@@ -402,6 +428,11 @@ def do_proxy_request(proxy, url, params={}, referer=None, ret_json=True, client=
         request_method = get_request_method(proxied_url, method=method, client=client)
     if referer:
         headers[enums.headers.referer] = referer
+
+    if request_method.__name__ == 'get':
+        # must use kwargs after url in GET
+        return request_method(proxied_url, params=params, headers=headers)
+
     return request_method(proxied_url, params, headers=headers)
 
 
@@ -409,7 +440,7 @@ def guess_proxy_url(domain):
     """Grade school level hack to see if there is a standard esri proxy available
             for a domain.
 
-    Arg:
+    Args:
         domain: URL to domain to check for proxy.
 
     Returns:
@@ -444,13 +475,14 @@ def guess_proxy_url(domain):
 def validate_name(file_name):
     """Validates an output name by removing special characters.
 
-    Arg:
+    Args:
         file_name: The name of the file to be validated.
 
     Returns:
         The path for the file.
     """
 
+
     import string
     path = os.sep.join(file_name.split(os.sep)[:-1]) #forward slash in name messes up os.path.split()
     name = fix_encoding(file_name.split(os.sep)[-1])
@@ -463,7 +495,7 @@ def validate_name(file_name):
 def guess_wkid(wkt):
     """Attempts to guess a well-known ID from a well-known text imput (WKT).
 
-    Arg:
+    Args:
         wkt: Well known text spatial reference
 
     Returns:
@@ -484,7 +516,7 @@ def guess_wkid(wkt):
 def assign_unique_name(fl):
     """Assigns a unique file name.
 
-    Arg:
+    Args:
         fl: Path of file.
 
     Returns:
@@ -506,7 +538,7 @@ def mil_to_date(mil):
     """Date items from REST services are reported in milliseconds,
             this function will convert milliseconds to datetime objects.
 
-    Arg:
+    Args:
         mil: Time in milliseconds.
 
     Returns:
@@ -520,10 +552,8 @@ def mil_to_date(mil):
     elif mil < 0:
         return datetime.datetime.utcfromtimestamp(0) + datetime.timedelta(seconds=(mil/1000))
     else:
-        # safely cast, to avoid being out of range for platform local time
         try:
-            struct = time.gmtime(mil /1000.0)
-            return datetime.datetime.fromtimestamp(time.mktime(struct))
+            return datetime.datetime.utcfromtimestamp(mil / 1000)
         except Exception as e:
             warnings.warn('bad milliseconds value: {}'.format(mil))
             raise e
@@ -531,7 +561,7 @@ def mil_to_date(mil):
 def date_to_mil(date=None):
     """Converts datetime.datetime() object to milliseconds.
 
-    Arg:
+    Args:
         date: datetime.datetime() object
 
     Returns:
@@ -545,7 +575,7 @@ def date_to_mil(date=None):
 def fix_encoding(s):
     """Fixes unicode by treating as ascii and ignoring errors.
 
-    Arg:
+    Args:
         s: Unicode string.
     """
 
@@ -569,7 +599,7 @@ def generate_token(url, user, pw, expiration=60, client=None, **kwargs):
     Returns:
         The token for security.
     """
-
+    ID_MANAGER.flush()
     suffix = '/rest/info'
     isAdmin = False
     if '/admin/' in url:
@@ -610,7 +640,7 @@ def generate_token(url, user, pw, expiration=60, client=None, **kwargs):
               USER_NAME: user,
               PASSWORD: pw,
               CLIENT: REQUEST_IP,
-              EXPIRATION: max([expiration, shortLived])}
+              EXPIRATION: min([expiration, shortLived])}
 
     # headers = {}
     if is_agol:
@@ -681,13 +711,20 @@ def generate_token(url, user, pw, expiration=60, client=None, **kwargs):
     if is_agol:
         if isinstance(org_resp, dict):
             org_id = org_resp.get('id')
+            base_url, url_key = org_resp.get('customBaseUrl'), org_resp.get('urlKey')
             if org_id:
-                # print('setting portal token: https://services2.arcgis.com/{}/arcgis/rest/services'.format(org_id))
                 token_copy = munch.munchify({})
                 token_copy.update(token.json)
                 serv_url = infoUrl.replace('/info', '/services')
                 token_copy.domain = serv_url
                 ID_MANAGER.tokens[serv_url] = Token(token_copy)
+            if base_url and url_key:
+                token_copy = munch.munchify({})
+                token_copy.update(token.json)
+                serv_url = get_portal_base('{}://{}.{}'.format(PROTOCOL, url_key, base_url))
+                token_copy.domain = serv_url
+                ID_MANAGER.tokens[serv_url] = Token(token_copy)
+
 
     if is_portal:
         if isinstance(portal_resp, dict):
@@ -730,7 +767,7 @@ def generate_elevated_portal_token(server_url, user_token, client=None, **kwargs
     Returns:
         The elevated portal token.
     """
-
+    ID_MANAGER.flush()
     params = {
         TOKEN: str(user_token) if isinstance(user_token, Token) else user_token,
         "serverURL": server_url,
@@ -760,12 +797,13 @@ def generate_elevated_portal_token(server_url, user_token, client=None, **kwargs
     ID_MANAGER.tokens[token.domain] = token
     return token
 
+
 class NameEncoder(json.JSONEncoder):
     """encoder for restapi objects to make serializeable for JSON"""
     def default(self, o):
         """Encodes object for JSON.
 
-        Arg:
+        Args:
             o: Object.
         """
         return o.__repr__()
@@ -777,7 +815,7 @@ class JsonGetter(object):
     def get(self, name, default=None):
         """Gets an attribute from json.
 
-        Arg:
+        Args:
             name: Name of attribute.
         """
         return self.json.get(name, default)
@@ -949,7 +987,7 @@ class RESTEndpoint(JsonGetter):
                 service is compatible with the version if it is greater than or
                 equal to the input version.
 
-        Arg:
+        Args:
             version: Minimum version compatibility as float (ex: 10.3 or 10.31).
 
         Returns:
@@ -973,7 +1011,8 @@ class RESTEndpoint(JsonGetter):
 
     def request(self, *args, **kwargs):
         """Wrapper for request to automatically pass in credentials."""
-        for key, value in six.iteritems({'token': 'token',
+        for key, value in six.iteritems({
+            'token': 'token',
             'cookies': '_cookie',
             'proxy': '_proxy',
             'referer': '_referer'
@@ -983,8 +1022,7 @@ class RESTEndpoint(JsonGetter):
 
         if 'ret_json' not in kwargs:
             kwargs['ret_json'] = True
-        if kwargs.get('stream'):
-            kwargs['ret_json'] = False
+
         kwargs['client'] = self.client
         return do_request(*args, **kwargs)
 
@@ -1033,6 +1071,10 @@ class SpatialReferenceMixin(object):
         if not isinstance(in_json, dict):
             return None
         for k, v in six.iteritems(in_json):
+            if k == CRS:
+                prop = in_json.get(CRS, {}).get(PROPERTIES, {}).get(NAME, '')
+                if ('ESPG:') in prop:
+                    return int(prop.replace('ESPG:',''))
             if k == SPATIAL_REFERENCE:
                 if isinstance(v, int):
                     return v
@@ -1102,6 +1144,19 @@ class SpatialReferenceMixin(object):
         """Returns the well known text (if it exists) for a service."""
         return self._spatialReference.get(WKT, '')
 
+    def getCRS(self):
+        """returs the crs representation if WKID exists"""
+        wkid = self.getWKID()
+        if wkid:
+            return munch.munchify({
+                TYPE: NAME,
+                PROPERTIES: {
+                    NAME: 'ESPG:{}'.format(wkid)
+                }
+            })
+
+        return None
+
 
 class FieldsMixin(object):
     json = {}
@@ -1147,6 +1202,8 @@ class FieldsMixin(object):
 
 class FeatureSetBase(JsonGetter, SpatialReferenceMixin, FieldsMixin):
     """Base Class for feature set."""
+    _format = None
+
     @property
     def hasGeometry(self):
         """Returns for if it has geometry."""
@@ -1162,13 +1219,11 @@ class FeatureSetBase(JsonGetter, SpatialReferenceMixin, FieldsMixin):
 
     def __getitem__(self, key):
         """Supports grabbing feature by index and json keys by name."""
-        try:
-            if isinstance(key, int):
-                return Feature(self.json.features[key])
-            else:
-                return Feature(self.json.get(key))
-        except:
-            return self.json.get(key)
+
+        if isinstance(key, int):
+            return Feature(self.json.features[key])
+
+        return self.json.get(key)
 
     def __iter__(self):
         for feature in self.features:
@@ -1194,7 +1249,7 @@ class FeatureSet(FeatureSetBase):
     def __init__(self, in_json):
         """Inits Class with input JSON for feature set.
 
-        Arg:
+        Args:
             in_json: Input json response from request.
 
         Raises:
@@ -1218,7 +1273,7 @@ class FeatureSet(FeatureSetBase):
     def extend(self, other):
         """Combines features from another FeatureSet with this one.
 
-        Arg:
+        Args:
             other: Other FeatureSet to combine with this one.
         """
         if not isinstance(other, FeatureSet):
@@ -1248,14 +1303,14 @@ class FeatureSet(FeatureSetBase):
         return FeatureSet(fsd)
 
 
-class GeoJSONFeatureSet(FeatureSetBase):
-    """Class that handles Geo JSON feature set."""
+class FeatureCollection(FeatureSetBase):
+    """Class that handles Geo JSON formatted Feature Set, known as a FeatureCollection."""
     _format = GEOJSON_FORMAT
 
     def __init__(self, in_json):
         """Inits class with JSON as geo feature set.
 
-        Arg:
+        Args:
             in_json: Input json response from request.
         """
 
@@ -1274,12 +1329,12 @@ class GeoJSONFeatureSet(FeatureSetBase):
     def extend(other):
         """Combines features from another FeatureSet with this one.
 
-        Arg:
+        Args:
             other: Other FeatureSet to combine with this one.
         """
 
-        if not isinstance(other, GeoJSONFeatureSet):
-            other = GeoJSONFeatureSet(other)
+        if not isinstance(other, FeatureCollection):
+            other = FeatureCollection(other)
         otherCopy = copy.deepcopy(other)
 
         # get max oid
@@ -1300,7 +1355,7 @@ class GeoJSONFeatureSet(FeatureSetBase):
             if k != FEATURES:
                 fsd[k] = v
         fsd[FEATURES] = []
-        return GeoJSONFeatureSet(fsd)
+        return FeatureCollection(fsd)
 
 
 class Feature(JsonGetter):
@@ -1308,7 +1363,7 @@ class Feature(JsonGetter):
     def __init__(self, feature):
         """Inits the class with a feature.
 
-        Arg:
+        Args:
             feature: Input json for feature.
         """
 
@@ -1319,7 +1374,7 @@ class Feature(JsonGetter):
     def get(self, field, default=None):
         """Returns/gets an attribute from the feature.
 
-        Arg:
+        Args:
             field: Name of field for which to get attribute.
         """
 
@@ -1344,7 +1399,7 @@ class RelatedRecords(JsonGetter, SpatialReferenceMixin):
     def __init__(self, in_json):
         """Inits class with json for query related records.
 
-        Arg:
+        Args:
             in_json: json response for query related records operation.
         """
 
@@ -1359,7 +1414,7 @@ class RelatedRecords(JsonGetter, SpatialReferenceMixin):
     def get_related_records(self, oid):
         """Gets the related records for an object id.
 
-        Arg:
+        Args:
             oid: Object ID for related records.
 
         Returns:
@@ -1400,13 +1455,13 @@ class BaseService(RESTEndpoint, SpatialReferenceMixin):
             self.name = self.url.split('/')[-2]
         self.name = self.name.split('/')[-1]
 
+    @property
+    def servicePath(self):
+        return self.url.split('/rest/services/')[-1]
+
     def __repr__(self):
         """String representation with service name."""
-        try:
-            qualified_name = '/'.join(filter(None, [self.url.split('/services/')[-1].split('/' + self.name)[0], self.name]))
-        except:
-            qualified_name = self.name
-        return '<{}: {}>'.format(self.__class__.__name__, qualified_name)
+        return '<{}: {}>'.format(self.__class__.__name__, self.servicePath)
 
 class OrderedDict2(OrderedDict):
     """Wrapper for OrderedDict."""
@@ -1420,7 +1475,7 @@ class PortalInfo(JsonGetter):
     def __init__(self, response):
         """Inits class with response from server.
 
-        Arg:
+        Args:
             response: The response from server.
         """
 
@@ -1486,7 +1541,8 @@ class Token(JsonGetter):
     @property
     def isExpired(self):
         """Boolean value for expired or not."""
-        if datetime.datetime.now() > self.time_expires:
+        now = datetime.datetime.utcnow()
+        if now > self.time_expires:
             return True
         else:
             return False
@@ -1723,28 +1779,19 @@ class EditResult(JsonGetter):
             print('Updated {} feature(s)'.format(self.success_count(getattr(self, UPDATE_RESULTS))))
         if self.json.get(DELETE_RESULTS, []):
             print('Deleted {} feature(s)'.format(self.success_count(getattr(self, DELETE_RESULTS))))
-        if self.json.get(ATTACHMENTS, []):
-            print('Attachment Edits: {}'.format(self.success_count(getattr(self, ATTACHMENTS))))
-        if self.json.get(ADD_ATTACHMENT_RESULT):
-            try:
-                k,v = list(getattr(self, ADD_ATTACHMENT_RESULT).items())[0]
-                print("Added attachment '{}' for feature {}".format(v, k))
-            except IndexError: # should never happen?
-                print('Added 1 attachment')
-        if self.json.get(DELETE_ATTACHMENT_RESULTS):
-            try:
-                for res in getattr(self, DELETE_ATTACHMENT_RESULTS, []) or []:
-                    if res.get(SUCCESS_STATUS) in (True, TRUE):
-                        print("Deleted attachment '{}'".format(res.get(RESULT_OBJECT_ID)))
-                    else:
-                        print("Failed to Delete attachment '{}'".format(res.get(RESULT_OBJECT_ID)))
-            except IndexError: # should never happen?
-                print('Deleted {} attachment(s)'.format(len(getattr(self, DELETE_ATTACHMENT_RESULT))))
-        if self.json.get(UPDATE_ATTACHMENT_RESULT):
-            try:
-                print("Updated attachment '{}'".format(self.json.get(UPDATE_ATTACHMENT_RESULT, {}).get(RESULT_OBJECT_ID)))
-            except IndexError: # should never happen?
-                print('Updated 1 attachment')
+        attResults = self.json.get(ATTACHMENTS)
+        if attResults:
+            for resAttr in (ADD_RESULTS, UPDATE_RESULTS):
+                results = attResults.get(resAttr, [])
+                if results:
+                    print('Attachment {} operation successful for {} of {} attachment(s)'.format(
+                        resAttr.replace('Results', ''),
+                        self.success_count(attResults.get(resAttr)),
+                        len(results)
+                    ))
+            dels = attResults.get(DELETE_RESULTS)
+            if dels:
+                print('Successfully deleted {} attachments')
 
 class BaseGeometry(SpatialReferenceMixin):
     """Base geometry obect."""
@@ -1792,10 +1839,9 @@ class GeocodeService(RESTEndpoint):
     """Class to handle Geocode Service."""
 
     def geocodeAddresses(self, recs, outSR=4326, address_field=''):
-        """Geocodes a list of addresses.  If there is a singleLineAddress field
-                present in the geocoding service, the only input required is a
-                list of addresses.  Otherwise, a record set an be passed in for
-                the "recs" parameter.  See formatting example at bottom.
+        """Geocodes a list of addresses.  If there is a singleLineAddress field present in the
+        geocoding service, the only input required is a list of addresses.  Otherwise, a record
+        set an be passed in for the "recs" parameter.  See formatting example at bottom.
 
         Args:
             recs: JSON object for fields as record set if no SingleLine field
@@ -1804,9 +1850,8 @@ class GeocodeService(RESTEndpoint):
             outSR: Optional output spatial refrence for geocoded addresses.
             address_field: Name of address field or Single Line address field.
 
-        # recs param examples
-        # preferred option as record set (from esri help docs):
-        recs = {
+        >>> # preferred option as record set (from esri help docs):
+        >>> recs = {
             "records": [
                 {
                     "attributes": {
@@ -1824,8 +1869,8 @@ class GeocodeService(RESTEndpoint):
                 }
             ]
         }
-        # full address list option if singleLineAddressField is present
-        recs = ['100 S Riverfront St, Mankato, MN 56001',..]
+        >>> # full address list option if singleLineAddressField is present
+        >>> recs = ['100 S Riverfront St, Mankato, MN 56001',..]
 
         Raises:
             ValueError: 'Not a valid input for "recs" parameter!'
